/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam;

import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.Enumeration;

public class CertificateLoader {

    // Suppress default constructor for noninstantiability
    private CertificateLoader() {
        throw new AssertionError();
    }

    public static X509Certificate loadCertificate(String certFile, String password) throws CertificateException, KeyStoreException, NoSuchAlgorithmException, IOException {
        KeyStore ks = loadKeyStore(certFile, password);
        return (X509Certificate) ks.getCertificate(aliasFromKeystore(ks));
    }

    public static RSAPrivateKey loadPrivateKey(String certFile, String password) throws CertificateException, UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException, IOException {
        KeyStore ks = loadKeyStore(certFile, password);
        return (RSAPrivateKey) ks.getKey(aliasFromKeystore(ks), password.toCharArray());
    }

    private static KeyStore loadKeyStore(String certFile, String password) throws CertificateException, KeyStoreException, NoSuchAlgorithmException, IOException {
        InputStream certStream = CertificateLoader.class.getResourceAsStream("/" + certFile);
        String type = typeFromSuffix(certFile);
        return loadKeyStore(certStream, type, password);
    }

    private static KeyStore loadKeyStore(InputStream certStream, String type, String password) throws CertificateException, KeyStoreException, NoSuchAlgorithmException, IOException {
        KeyStore ks = KeyStore.getInstance(type);
        ks.load(certStream, password.toCharArray());
        return ks;
    }

    private static String typeFromSuffix(String path) {
        return path.toLowerCase().endsWith(".p12") ? "PKCS12" : "JKS";
    }

    private static String aliasFromKeystore(KeyStore ks) throws KeyStoreException {
        String alias = null;
        Enumeration<String> e = ks.aliases();
        while (e.hasMoreElements()) {
            if (alias != null) {
                throw new IllegalStateException("Multiple alias in keystore");
            }
            alias = e.nextElement();
        }
        return alias;
    }


}
