/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam;

import com.google.common.base.MoreObjects;
import com.google.common.base.Splitter;
import com.ibm.security.access.extension.authn.AuthenticationMechanism;
import com.ibm.security.access.extension.authn.AuthenticationMechanismContext;
import com.ibm.security.access.extension.authn.AuthenticationMechanismException;
import com.ibm.security.access.extension.authn.AuthenticationMechanismResult;
import com.ibm.security.access.extension.authn.AuthenticationMechanismResultAbortTask;
import com.ibm.security.access.extension.authn.AuthenticationMechanismResultPauseTask;
import com.ibm.security.access.extension.authn.credential.Attribute;
import com.ibm.security.access.extension.authn.task.HtmlPageAbortTask;
import com.ibm.security.access.extension.authn.task.HtmlPagePauseTask;
import no.buypass.code.api.isam.http.DefaultOtpValidator;
import no.buypass.code.api.isam.ldap.DefaultLdapLookup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.InetAddress;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.Collections;
import java.util.List;
import java.util.Properties;


public class CodeApiAuthenticationMechanism implements AuthenticationMechanism {

    private static final Logger LOGGER = LoggerFactory.getLogger(CodeApiAuthenticationMechanism.class);

    /* static attributes */
    private static final String AUTH_MECH_TYPES = "authenticationMechanismTypes";
    private static final String CODE_HTML = "/authsvc/authenticator/buypass_code_api/code.html";
    private static final String ERROR_HTML = "/authsvc/authenticator/buypass_code_api/error.html";

    /* private members */
    private String authenticatorId;
    private OtpValidator otpValidator;
    private LdapLookup ldapLookup;

    public CodeApiAuthenticationMechanism() {
    }

    @Override
    public AuthenticationMechanismResult execute(AuthenticationMechanismContext ctx)
            throws AuthenticationMechanismException {
        try {

            LOGGER.info("Execute Buypass Code API mechanism {}", ctx);

            String hostName = InetAddress.getLocalHost().getHostName();
            LOGGER.warn("Local host name {}", hostName);

            String address = InetAddress.getLocalHost().getHostAddress();
            LOGGER.warn("Local host address {}", address);


			/*
             * You're probably not logged in if this executes
			 */
            if (ctx.getCredential() == null) {
                LOGGER.info("Abort, no context credential");
                return AuthenticationMechanismResult.abort(failure("No context credential"));
            }

            String username = ctx.getCredential().getUsername();

			/* they havent authenticated */
            Attribute authMechTypes = ctx.getCredential().getAttribute(AUTH_MECH_TYPES);
            if (authMechTypes == null || !authMechTypes.hasValue(authenticatorId)) {

				/* Fetch an OTP */
                String otp = ctx.getRequest().getParameter("otp");
                if (otp == null || otp.length() <= 0) {
                    return AuthenticationMechanismResult.pause(getOTPTask());
                }

                Authenticator authenticator;
                try {
                    authenticator = ldapLookup.resolve(username);
                } catch (Exception e) {
                    return AuthenticationMechanismResult.abort(failure(e.getMessage()));
                }

                OtpValidation result = otpValidator.validate(username, authenticator, otp);
                if (result.isValid()) {
                    return AuthenticationMechanismResult.success();
                }
                else {
                    AuthenticationMechanismResultPauseTask task = getOTPTask(result.getReplyMessage());
                    return AuthenticationMechanismResult.pause(task);
                }
            }
            else {
                return AuthenticationMechanismResult.success();
            }
        } catch (Exception e) {
            StringWriter errors = new StringWriter();
            e.printStackTrace(new PrintWriter(errors));
            String stackTrace = errors.toString();
            String message = "Caught general exception: " + e + " " + stackTrace;
            LOGGER.error("Caught general exception " + e);
            return AuthenticationMechanismResult.abort(failure(message));
        }
    }

    private AuthenticationMechanismResultAbortTask failure(String errorMessage) {
        HtmlPageAbortTask failTask = new HtmlPageAbortTask();
        failTask.setPageId(ERROR_HTML);
        failTask.setMacro("@ERROR_MESSAGE@", errorMessage);
        return failTask;
    }

    private AuthenticationMechanismResultPauseTask getOTPTask() {
        HtmlPagePauseTask otpTask = new HtmlPagePauseTask();
        otpTask.setPageId(CODE_HTML);
        otpTask.setMacro("@REPLY_MESSAGE@", "");
        return otpTask;
    }

    private AuthenticationMechanismResultPauseTask getOTPTask(String replyMessage) {
        HtmlPagePauseTask otpTask = new HtmlPagePauseTask();
        otpTask.setPageId(CODE_HTML);
        otpTask.setMacro("@REPLY_MESSAGE@", replyMessage);
        return otpTask;
    }


    @Override
    public void shutdown() {
        LOGGER.info("Shutdown Buypass Code API mechanism ");
    }

    @Override
    public void startup(Properties prp) {
        LOGGER.info("Starting Buypass Code API mechanism");
        authenticatorId = prp.getProperty(PropertyKey.AUTH_ID);
        initCodeIdentityResolver(prp);
        initOtpValidator(prp);
        LOGGER.info("Buypass Code API mechanism started");
    }

    private void initCodeIdentityResolver(Properties prp) {
        String ldapProviderUrl = stringProperty(prp, PropertyKey.LDAP_PROVIDER_URL);
        String ldapSecurityPrincipal = stringProperty(prp, PropertyKey.LDAP_SECURITY_PRINCIPAL);
        String ldapSecurityCredentials = stringProperty(prp, PropertyKey.LDAP_SECURITY_CREDENTIALS);
        List<String> ldapPhoneNumberAttrNames = csvProperty(prp, PropertyKey.LDAP_PHONE_NUMBER_ATTR_NAME);
        List<String> ldapAppIdAttrNames = csvProperty(prp, PropertyKey.LDAP_APP_ID_ATTR_NAME);
        String ldapUserAttrName = stringProperty(prp, PropertyKey.LDAP_USER_ATTR_NAME);
        String ldapBaseDn = stringProperty(prp, PropertyKey.LDAP_BASE_DN);
        ldapLookup = new DefaultLdapLookup(ldapProviderUrl, ldapSecurityPrincipal, ldapSecurityCredentials, ldapPhoneNumberAttrNames, ldapAppIdAttrNames, ldapBaseDn, ldapUserAttrName);
    }

    private void initOtpValidator(Properties prp) {

        String codeApiUrl = stringProperty(prp, PropertyKey.CODE_API_URL);
        String clientId = stringProperty(prp, PropertyKey.CLIENT_ID);
        String keystorePassword = stringProperty(prp, PropertyKey.KEYSTORE_PASSWORD);
        String keystoreLocation = stringProperty(prp, PropertyKey.KEYSTORE_NAME);
        String proxyUrl = stringProperty(prp, PropertyKey.PROXY_URL);
        String proxyUsername = stringProperty(prp, PropertyKey.PROXY_USERNAME);
        String proxyPassword = stringProperty(prp, PropertyKey.PROXY_PASSWORD);

        try {
            RSAPrivateKey clientPrivateKey = CertificateLoader.loadPrivateKey(keystoreLocation, keystorePassword);
            X509Certificate clientCertificate = CertificateLoader.loadCertificate(keystoreLocation, keystorePassword);

            otpValidator = DefaultOtpValidator.builder()
                    .serverUrl(codeApiUrl)
                    .clientId(clientId)
                    .clientCertificate(clientCertificate)
                    .clientPrivateKey(clientPrivateKey)
                    .proxyUrl(proxyUrl)
                    .proxyUsername(proxyUsername)
                    .proxyPassword(proxyPassword)
                    .build();
        } catch (Exception e) {
            throw new IllegalStateException("Failed to initialize authentication mechanism", e);
        }
    }

    private List<String> csvProperty(Properties prp, String key) {
        String property = prp.getProperty(key);
        return property != null ? Splitter.on(',').splitToList(property) : Collections.emptyList();
    }

    private String stringProperty(Properties prp, String key) {
        String property = prp.getProperty(key);
        return property != null ? property.trim() : null;
    }

    @Override
    public String toString() {
        return MoreObjects.toStringHelper(this)
                .add("authenticatorId", authenticatorId)
                .add("otpValidator", otpValidator)
                .toString();
    }
}
