/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam.http;

import no.buypass.api.code.authentication.client.AuthenticationClient;
import no.buypass.api.code.authentication.client.model.AccessTokenResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.security.AccessControlException;
import java.time.Clock;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class ClientCredentials {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientCredentials.class);

    /**
     * Clock used for expiration checks.
     */
    private final Clock clock;

    private final AuthenticationClient authenticationClient;

    /**
     * Lock on the token response information.
     */
    private final Lock lock = new ReentrantLock();

    private String tokenType;

    private String accessToken;

    private Long expirationTimeMilliseconds;

    public ClientCredentials(AuthenticationClient authenticationClient) {
        this(authenticationClient, Clock.systemUTC());
    }

    public ClientCredentials(AuthenticationClient authenticationClient, Clock clock) {
        this.authenticationClient = authenticationClient;
        this.clock = clock;
    }

    public String getTokenType() {
        return tokenType != null ? tokenType : "Bearer";
    }

    public String getAccessToken() {
        lock.lock();
        try {
            Long expiresIn = getExpiresInSeconds();
            // check if token will expire in a minute
            if (accessToken == null || expiresIn != null && expiresIn <= 60) {
                AccessTokenResponse accessTokenResponse = requestAccessToken();
                setFromTokenResponse(accessTokenResponse);
            }
            return accessToken;
        }
        finally {
            lock.unlock();
        }
    }

    private AccessTokenResponse requestAccessToken() {
        LOGGER.info("Request access token");

        // An AccessControlException is expected on the first invocation.
        try {
            return authenticationClient.requestAccessToken();
        } catch (AccessControlException e) {
            return authenticationClient.requestAccessToken();
        }
    }


    private void setFromTokenResponse(AccessTokenResponse accessTokenResponse) {
        lock.lock();
        try {
            this.accessToken = accessTokenResponse.getAccessToken();
            this.tokenType = accessTokenResponse.getTokenType();
            this.expirationTimeMilliseconds = System.currentTimeMillis() + accessTokenResponse.getExpiresInSeconds() * 1000L;
        }
        finally {
            lock.unlock();
        }
    }

    /**
     * Returns the remaining lifetime in seconds of the access token (for example 3600 for an hour, or
     * -3600 if expired an hour ago) or {@code null} if unknown.
     */
    private Long getExpiresInSeconds() {
        lock.lock();
        try {
            if (expirationTimeMilliseconds == null) {
                return null;
            }
            return (expirationTimeMilliseconds - clock.millis()) / 1000;
        }
        finally {
            lock.unlock();
        }
    }

}
