/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam.http;

import no.buypass.api.code.authentication.client.AuthenticationClient;
import no.buypass.api.code.authentication.client.DefaultAuthenticationClient;
import no.buypass.api.code.authentication.client.exception.AuthenticationClientException;
import no.buypass.api.code.authentication.client.model.UserAuthentication;
import no.buypass.api.code.authentication.client.model.UserCredentials;
import no.buypass.code.api.isam.Authenticator;
import no.buypass.code.api.isam.OtpValidation;
import no.buypass.code.api.isam.OtpValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.UUID;


public class DefaultOtpValidator implements OtpValidator {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultOtpValidator.class);

    private final AuthenticationClient authenticationClient;

    private final ClientCredentials clientCredentials;

    private DefaultOtpValidator(Builder builder) {

        this.authenticationClient = DefaultAuthenticationClient.builder(builder.serverUrl,
                builder.clientId,
                builder.clientCertificate,
                builder.clientPrivateKey)
                .proxyUrl(builder.proxyUrl)
                .proxyUsername(builder.proxyUsername)
                .proxyPassword(builder.proxyPassword)
                .build();

        this.clientCredentials = new ClientCredentials(authenticationClient);
    }

    public static Builder builder() {
        return new Builder();
    }

    /**
     *
     */
    @Override
    public OtpValidation validate(String userName, Authenticator authenticator, String otp) {

        LOGGER.info("Validate OTP for user {} ", userName);

        String replyMessage;

        try {

            String accessToken = clientCredentials.getAccessToken();
            String tokenType = clientCredentials.getTokenType();

            String sessionId = UUID.randomUUID().toString();
            UserCredentials userCredentials = UserCredentials.builder()
                    .userId(authenticator.getUserId())
                    .appId(authenticator.getAppId())
                    .secret(otp)
                    .sessionId(sessionId)
                    .build();

            UserAuthentication authentication = authenticationClient.authenticateUser(userName,
                    userCredentials,
                    tokenType,
                    accessToken);

            boolean validated = authentication.isAuthenticated();
            if (!validated) {
                LOGGER.info(userName + " authentication failed, " + authentication.getReplyMessage());
            }
            return OtpValidation.builder()
                    .valid(validated)
                    .replyMessage(authentication.getReplyMessage())
                    .build();
        } catch (AuthenticationClientException e) {
            replyMessage = e.getMessage();
            LOGGER.warn("Server error: " + replyMessage);
        } catch (Exception e) {
            StringWriter errors = new StringWriter();
            e.printStackTrace(new PrintWriter(errors));
            String stackTrace = errors.toString();
            e.printStackTrace();
            replyMessage = "Caught general exception: " + e + " " + stackTrace;
            LOGGER.warn("Caught general exception: " + e.getMessage());
        }
        return OtpValidation.builder()
                .valid(false)
                .replyMessage(replyMessage)
                .build();
    }

    public static final class Builder {

        private String serverUrl;
        private String clientId;
        private X509Certificate clientCertificate;
        private RSAPrivateKey clientPrivateKey;
        private String proxyUrl;
        private String proxyUsername;
        private String proxyPassword;

        private Builder() {
        }

        public Builder serverUrl(String val) {
            serverUrl = val;
            return this;
        }

        public Builder clientId(String val) {
            clientId = val;
            return this;
        }

        public Builder clientCertificate(X509Certificate val) {
            clientCertificate = val;
            return this;
        }

        public Builder clientPrivateKey(RSAPrivateKey val) {
            clientPrivateKey = val;
            return this;
        }

        public Builder proxyUrl(String val) {
            proxyUrl = val;
            return this;
        }

        public Builder proxyUsername(String val) {
            proxyUsername = val;
            return this;
        }

        public Builder proxyPassword(String val) {
            proxyPassword = val;
            return this;
        }

        public DefaultOtpValidator build() {
            return new DefaultOtpValidator(this);
        }
    }
}
