/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam.ldap;

import no.buypass.code.api.isam.Authenticator;
import no.buypass.code.api.isam.LdapLookup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;

public class DefaultLdapLookup implements LdapLookup {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultLdapLookup.class);

    //    private static final int NANO_PERIODS_IN_MILLIS = 10000;
    private static final int ACCOUNT_DISABLED = 0x0002;
    private static final int ACCOUNT_LOCKED_OUT = 0x00010;
    // 2**63 100ns windows periods in minutes
//    private static final long ACCOUNT_BLOCKED_FOREVER = 26113L;
//    private static final String LOCKOUT_DURATION_ATTRIBUTE = "getLockoutDuration";

    private static final String USER_ACCOUNT_CONTROL_ATTRIBUTE = "userAccountControl";
    private static final String LOCKOUT_TIME_ATTRIBUTE = "lockoutTime";


    private final Hashtable<String, String> env;

    private final List<String> ldapPhoneNumberAttrNames;
    private final List<String> ldapAppIdAttrNames;
    private final String ldapBaseDn;
    private final String ldapUserAttrName;

    private final int ldapTimeLimit;
    private final SearchControls searchControls;

    public DefaultLdapLookup(String ldapProviderUrl, String ldapSecurityPrincipal, String ldapSecurityCredentials, List<String> ldapPhoneNumberAttrNames, List<String> ldapAppIdAttrNames, String ldapBaseDn, String ldapUserAttrName) {
        ldapTimeLimit = 30_000;
        env = new Hashtable<>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
        env.put(Context.SECURITY_AUTHENTICATION, "simple");
        env.put(Context.PROVIDER_URL, ldapProviderUrl);
        env.put(Context.SECURITY_PRINCIPAL, ldapSecurityPrincipal);
        env.put(Context.SECURITY_CREDENTIALS, ldapSecurityCredentials);
        this.ldapPhoneNumberAttrNames = ldapPhoneNumberAttrNames;
        this.ldapAppIdAttrNames = ldapAppIdAttrNames;

        this.ldapBaseDn = ldapBaseDn;
        this.ldapUserAttrName = ldapUserAttrName;

        searchControls = searchControls();
    }

    @Override
    public Authenticator resolve(String userName) {

        try {
            SearchResult result = search(userName);
            Authenticator identifiers = resolveCodeIdentity(result.getAttributes().getAll());
            checkUserAccountDisabled(result.getAttributes().getAll());
            checkUserAccountLockedOut(result.getAttributes().getAll());
            return identifiers;
        } catch (NamingException e) {
            throw new IllegalStateException("LDAP exception", e);
        }
    }


    private SearchResult search(String userName) throws NamingException {

        DirContext ctx = new InitialDirContext(env);
        //String filter = "sn=Barnes";
        String filter = ldapUserAttrName + "=" + userName;

        LOGGER.info("LDAP search for {} in {}", filter, ldapBaseDn);

        NamingEnumeration<SearchResult> results = ctx.search(ldapBaseDn, filter, searchControls);
        SearchResult result = ensureSingleResult(results);
        if (result == null || result.getAttributes() == null) {
            throw new IllegalStateException("Empty search result for user " + userName);
        }
        return result;
    }

    private void checkUserAccountDisabled(NamingEnumeration<? extends Attribute> attributes) throws NamingException {
        String userAccountControlAttribute = getFirstAttributeValue(attributes, USER_ACCOUNT_CONTROL_ATTRIBUTE);
        if (userAccountControlAttribute != null && !userAccountControlAttribute.isEmpty()) {
            int value = Integer.parseInt(userAccountControlAttribute);
            boolean disabled = (value & ACCOUNT_DISABLED) == ACCOUNT_DISABLED;
            if (disabled) {
                throw new IllegalStateException("The user account is disabled");
            }
        }
    }

    private void checkUserAccountLockedOut(NamingEnumeration<? extends Attribute> attributes) throws NamingException {
        String userAccountControlAttribute = getFirstAttributeValue(attributes, LOCKOUT_TIME_ATTRIBUTE);
        if (userAccountControlAttribute != null && !userAccountControlAttribute.isEmpty()) {
            int value = Integer.parseInt(userAccountControlAttribute);
            boolean lockedOut = (value & ACCOUNT_LOCKED_OUT) == ACCOUNT_LOCKED_OUT;
            if (lockedOut) {
                throw new IllegalStateException("The account is currently locked out");
            }
        }

//        The above is not sufficient for detecting lockedOut. There is an alternative method, but this is not
//        implemented now because it is difficult to test.

//        Not set or 0 in lockoutTime is OK
//        LOGGER.trace("Now checking attribute {}", LOCKOUT_TIME_ATTRIBUTE);
//
//        String lockoutTime = getFirstAttributeValue(attributes, LOCKOUT_TIME_ATTRIBUTE);
//        if ((lockoutTime == null) || lockoutTime.isEmpty() || is0(lockoutTime)) {
//
//        }
//
//        Duration duration = getLockoutDuration(ldapBaseDn);
    }

    private SearchResult ensureSingleResult(NamingEnumeration<SearchResult> result) throws NamingException {
        if (result.hasMore()) {
            SearchResult searchResult = result.next();
            if (result.hasMore()) {
                throw new IllegalStateException("Search should return a single row, please check your search criteria");
            }
            return searchResult;
        }
        return null;
    }

    private Authenticator resolveCodeIdentity(NamingEnumeration<? extends Attribute> attributes) throws NamingException {
        String userId = getFirstAttributeValue(attributes, ldapPhoneNumberAttrNames);
        String appId = getFirstAttributeValue(attributes, ldapAppIdAttrNames);
        return Authenticator.builder()
                .userId(userId)
                .appId(appId)
                .build();
    }

    private String getFirstAttributeValue(NamingEnumeration<? extends Attribute> attributes, String... attributeNames) throws NamingException {
        List<String> names = Arrays.asList(attributeNames);
        return getFirstAttributeValue(attributes, names);
    }

    private String getFirstAttributeValue(NamingEnumeration<? extends Attribute> attributes, List<String> attributeNames) throws NamingException {
        while (attributes.hasMore()) {
            Attribute attr = attributes.next();
            //LOG.trace("Read attribute {}", attr);
            if (isOneOf(attr, attributeNames)) {
                return attrValue(attr);
            }
        }
        return null;
    }

    private boolean isOneOf(Attribute attr, List<String> candidateKeys) {
        for (String candidateKey : candidateKeys) {
            if (candidateKey.equalsIgnoreCase(attrName(attr))) {
                //LOG.trace("Found {} for {} in  {}", attrValue(attr), type, candidateKey);
                return true;
            }
        }
        return false;
    }

    private static String attrName(Attribute attr) {
        return attr != null ? attr.getID().toLowerCase() : null;
    }

    private static String attrValue(Attribute attr) {
        try {
            return attr != null && attr.get() != null ? attr.get().toString() : null;
        } catch (NamingException e) {
            return null;
        }
    }

    private SearchControls searchControls() {
        SearchControls searchControls = new SearchControls();
        searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
        searchControls.setTimeLimit(ldapTimeLimit);
        return searchControls;
    }

//    private Duration getLockoutDuration(String baseDn) {
//        try {
//            String name = policyDn(baseDn);
//            DirContext ctx = new InitialDirContext(env);
//            String lockoutDuration = ctx.getAttributes(name).get(LOCKOUT_DURATION_ATTRIBUTE).get().toString();
//            Duration dur = toDuration(lockoutDuration);
//            LOGGER.trace("Lockout duration from policy object is {}", dur);
//            return dur;
//        }
//        catch (Exception e) {
//            LOGGER.warn("Unable to read policy object, is this really AD? Reverting to config lockout duration value", e);
//            return null;
//        }
//    }

//    private static boolean is0(String lockoutTicks) {
//        try {
//            return (lockoutTicks == null) || (Integer.valueOf(lockoutTicks) == 0);
//        }
//        catch (NumberFormatException e) {
//            return false;
//        }
//    }

//    private static String policyDn(String baseDN) {
//        List<String> rdns = Splitter.on(',').splitToList(baseDN);
//        if (rdns.isEmpty()) {
//            throw new IllegalStateException("Malformed Base DN " + baseDN);
//        }
//
//        List<String> dcs = new ArrayList<>();
//        for (String rdn : rdns) {
//            if (rdn.toLowerCase().startsWith("policyDn=")) {
//                dcs.add(rdn);
//            }
//        }
//        if (dcs.isEmpty()) {
//            throw new IllegalStateException("No DC elements in base DN " + baseDN);
//        }
//
//        String dc = Joiner.on(',').join(dcs).trim();
//        LOGGER.trace("Will look for policy object in {}", dc);
//        return dc;
//    }

//    private static Duration toDuration(String value) {
//        BigInteger[] milliseconds = new BigInteger(value)
//                .divideAndRemainder(BigInteger.valueOf(NANO_PERIODS_IN_MILLIS));
//        BigInteger[] seconds = milliseconds[0].divideAndRemainder(BigInteger
//                .valueOf(DateTimeConstants.MILLIS_PER_SECOND));
//        BigInteger[] minutes = seconds[0].divideAndRemainder(BigInteger.valueOf(DateTimeConstants.SECONDS_PER_MINUTE));
//
//        long millis = minutes[0].intValue() * DateTimeConstants.MILLIS_PER_SECOND
//                * DateTimeConstants.SECONDS_PER_MINUTE;
//        millis += minutes[1].intValue() * DateTimeConstants.MILLIS_PER_SECOND;
//        return Duration.ofMillis(millis);
//    }
}
