/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.code.api.isam;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.junit.WireMockRule;
import com.ibm.security.access.extension.authn.AuthenticationMechanism;
import com.ibm.security.access.extension.authn.AuthenticationMechanismContext;
import com.ibm.security.access.extension.authn.AuthenticationMechanismException;
import com.ibm.security.access.extension.authn.AuthenticationMechanismResult;
import com.ibm.security.access.extension.authn.AuthenticationMechanismResultType;
import com.ibm.security.access.extension.authn.credential.Credential;
import com.ibm.security.access.extension.authn.message.Request;
import com.unboundid.ldap.listener.InMemoryDirectoryServer;
import com.unboundid.ldap.listener.InMemoryDirectoryServerConfig;
import com.unboundid.ldif.LDIFReader;
import no.buypass.api.code.authentication.client.model.AccessTokenResponse;
import no.buypass.api.code.authentication.client.model.ErrorResponse;
import no.buypass.api.code.authentication.client.model.UserAuthentication;
import no.buypass.code.api.isam.extension.IsamSecurityManager;
import no.buypass.code.api.isam.extension.TestAuthenticationMechanismContext;
import no.buypass.code.api.isam.extension.TestRequest;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;

import javax.ws.rs.core.MediaType;
import java.io.InputStream;
import java.util.Properties;
import java.util.UUID;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static com.github.tomakehurst.wiremock.core.WireMockConfiguration.wireMockConfig;
import static org.assertj.core.api.Assertions.assertThat;

/**
 *
 */
public class CodeApiAuthenticationMechanismTest {

    private static final String LDIF_FILE_PATH = "/ldif/test-server.ldif";

    private InMemoryDirectoryServer directoryServer;

    private SecurityManager defaultSecurityManager;

    /**
     * Make mocks of LDAP Server and Code API
     */


    @Rule
    public final WireMockRule wireMockRule = new WireMockRule(wireMockConfig().dynamicPort());


    @Before
    public void setup() throws Exception {

        defaultSecurityManager = System.getSecurityManager();

        String baseDn1 = "ou=Groups,dc=example,dc=com";
        String baseDn2 = "ou=People,dc=example,dc=com";

        InMemoryDirectoryServerConfig config = new InMemoryDirectoryServerConfig(baseDn1, baseDn2);

        directoryServer = new InMemoryDirectoryServer(config);

        InputStream ldifStream = getClass().getResourceAsStream(LDIF_FILE_PATH);
        LDIFReader ldifReader = new LDIFReader(ldifStream);
        directoryServer.importFromLDIF(true, ldifReader);

        // start the listener
        directoryServer.startListening();



        System.setSecurityManager(new IsamSecurityManager());
    }

    @After
    public void tearDown() {
        System.setSecurityManager(defaultSecurityManager);
        if (directoryServer != null) {
            directoryServer.shutDown(true);
        }
    }

    @Test
    public void testExecute() throws AuthenticationMechanismException {

        setupStub();

        int port = directoryServer.getListenPort();
        String ldapUrl = "ldap://localhost:" + port;
        String codeApiUrl = "http://localhost:" + wireMockRule.port();
        String ldapSecurityPrincipal = "uid=admin,ou=People,dc=example,dc=com";
        String ldapSecurityCredentials = "admin";

        Properties p = new Properties();
        p.setProperty(PropertyKey.AUTH_ID, UUID.randomUUID().toString());
        p.setProperty(PropertyKey.LDAP_PROVIDER_URL, ldapUrl);
        p.setProperty(PropertyKey.LDAP_SECURITY_PRINCIPAL, ldapSecurityPrincipal);
        p.setProperty(PropertyKey.LDAP_SECURITY_CREDENTIALS, ldapSecurityCredentials);
        p.setProperty(PropertyKey.LDAP_PHONE_NUMBER_ATTR_NAME, "facsimiletelephonenumber");
        p.setProperty(PropertyKey.LDAP_BASE_DN, "ou=People,dc=example,dc=com");
        p.setProperty(PropertyKey.LDAP_USER_ATTR_NAME, "uid");

        p.setProperty(PropertyKey.CODE_API_URL, codeApiUrl);
        p.setProperty(PropertyKey.CLIENT_ID, "666669");
        p.setProperty(PropertyKey.KEYSTORE_PASSWORD, "zZhp2XcA");
        p.setProperty(PropertyKey.KEYSTORE_NAME, "cert/client-certificate.p12");

        AuthenticationMechanism mechanism = new CodeApiAuthenticationMechanism();
        mechanism.startup(p);

        AuthenticationMechanismResult result = mechanism.execute(createContext());
        assertThat(result.getType()).isEqualTo(AuthenticationMechanismResultType.SUCCESS);
    }

    @Test
    public void testExecuteFailure() throws AuthenticationMechanismException {

        setupErrorStub();

        int port = directoryServer.getListenPort();
        String ldapUrl = "ldap://localhost:" + port;
        String codeApiUrl = "http://localhost:" + wireMockRule.port();
        String ldapSecurityPrincipal = "uid=admin,ou=People,dc=example,dc=com";
        String ldapSecurityCredentials = "admin";

        Properties p = new Properties();
        p.setProperty(PropertyKey.AUTH_ID, UUID.randomUUID().toString());
        p.setProperty(PropertyKey.LDAP_PROVIDER_URL, ldapUrl);
        p.setProperty(PropertyKey.LDAP_SECURITY_PRINCIPAL, ldapSecurityPrincipal);
        p.setProperty(PropertyKey.LDAP_SECURITY_CREDENTIALS, ldapSecurityCredentials);
        p.setProperty(PropertyKey.LDAP_PHONE_NUMBER_ATTR_NAME, "facsimiletelephonenumber");
        p.setProperty(PropertyKey.LDAP_BASE_DN, "ou=People,dc=example,dc=com");
        p.setProperty(PropertyKey.LDAP_USER_ATTR_NAME, "uid");

        p.setProperty(PropertyKey.CODE_API_URL, codeApiUrl);
        p.setProperty(PropertyKey.CLIENT_ID, "666669");
        p.setProperty(PropertyKey.KEYSTORE_PASSWORD, "zZhp2XcA");
        p.setProperty(PropertyKey.KEYSTORE_NAME, "cert/client-certificate.p12");

        AuthenticationMechanism mechanism = new CodeApiAuthenticationMechanism();
        mechanism.startup(p);

        AuthenticationMechanismResult result = mechanism.execute(createContext());
        assertThat(result.getType()).isEqualTo(AuthenticationMechanismResultType.PAUSE);
    }

//    @Test
//    public void testExecuteProxy() throws AuthenticationMechanismException {
//        int port = directoryServer.getListenPort();
//        String ldapUrl = "ldap://localhost:" + port;
//        String codeApiUrl = "http://test12.bptest.local:8717/code-authentication-api";
//        String ldapSecurityPrincipal = "uid=admin,ou=People,dc=example,dc=com";
//        String ldapSecurityCredentials = "admin";
//
//        Properties p = new Properties();
//        p.setProperty(PropertyKey.AUTH_ID, UUID.randomUUID().toString());
//        p.setProperty(PropertyKey.LDAP_PROVIDER_URL, ldapUrl);
//        p.setProperty(PropertyKey.LDAP_SECURITY_PRINCIPAL, ldapSecurityPrincipal);
//        p.setProperty(PropertyKey.LDAP_SECURITY_CREDENTIALS, ldapSecurityCredentials);
//        p.setProperty(PropertyKey.LDAP_PHONE_NUMBER_ATTR_NAME, "facsimiletelephonenumber");
//        p.setProperty(PropertyKey.LDAP_BASE_DN, "ou=People,dc=example,dc=com");
//        p.setProperty(PropertyKey.LDAP_USER_ATTR_NAME, "uid");
//
//        p.setProperty(PropertyKey.CODE_API_URL, codeApiUrl);
//        p.setProperty(PropertyKey.CLIENT_ID, "666669");
//        p.setProperty(PropertyKey.KEYSTORE_PASSWORD, "zZhp2XcA");
//        p.setProperty(PropertyKey.KEYSTORE_NAME, "cert/client-certificate.p12");
//        p.setProperty(PropertyKey.PROXY_URL, "http://proxy.intra.buypass.no:8080");
//
//        AuthenticationMechanism mechanism = new CodeApiAuthenticationMechanism();
//        mechanism.startup(p);
//
//        AuthenticationMechanismResult result = mechanism.execute(createContext());
//        assertThat(result.getType()).isEqualTo(AuthenticationMechanismResultType.PAUSE);
//    }

    private AuthenticationMechanismContext createContext() {
        Credential credential = new Credential("abarnes");
        Request request = TestRequest.newBuilder().addParameter("otp", "123456").build();
        return TestAuthenticationMechanismContext.builder().credential(credential).request(request).build();
    }

    private void setupStub()  {

        try {
            ObjectMapper mapper = new ObjectMapper();

            AccessTokenResponse accessTokenResponse = AccessTokenResponse.builder()
                    .accessToken("sdafkjasdjifjsd")
                    .tokenType("Bearer")
                    .expiresInSeconds(120)
                    .build();

            UserAuthentication authentication = UserAuthentication.builder()
                    .authenticated(true)
                    .sessionId(UUID.randomUUID().toString())
                    .build();

            String accessTokenResponseBody = mapper.writeValueAsString(accessTokenResponse);
            String authenticationBody = mapper.writeValueAsString(authentication);

            givenThat(post(urlEqualTo("/auth/token"))
                    .willReturn(aResponse()
                            .withHeader("Content-Type", MediaType.APPLICATION_JSON)
                            .withStatus(200)
                            .withBody(accessTokenResponseBody)));

            givenThat(post(urlEqualTo("/user/abarnes/authenticate"))
                    .willReturn(aResponse()
                            .withHeader("Content-Type", MediaType.APPLICATION_JSON)
                            .withStatus(200)
                            .withBody(authenticationBody)));
        } catch (JsonProcessingException e) {
            throw new IllegalStateException("Failed to setup stub", e);
        }
    }

    private void setupErrorStub() {

        try {
            ObjectMapper mapper = new ObjectMapper();

            ErrorResponse errorResponse = ErrorResponse.builder()
                    .code(500)
                    .message("INVALID_TOKEN")
                    .details("Token has expired")
                    .build();

            String errorResponseBody = mapper.writeValueAsString(errorResponse);

            givenThat(post(urlEqualTo("/auth/token"))
                    .willReturn(aResponse()
                            .withHeader("Content-Type", MediaType.APPLICATION_JSON)
                            .withStatus(500)
                            .withBody(errorResponseBody)));
        } catch (JsonProcessingException e) {
            throw new IllegalStateException("Failed to setup stub", e);
        }

    }

}
